import boto3
from moto import mock_wafv2


@mock_wafv2
def test_list_rule_groups():
    client = boto3.client("wafv2", region_name="us-east-2")
    resp = client.list_rule_groups(Scope="CLOUDFRONT")
    resp.should.have.key("RuleGroups").equals([])
